// -*- C++ -*-
#include "Rivet/Analysis.hh"
#include "Rivet/Projections/FinalState.hh"

#include "Rivet/Particle.hh"
#include "Rivet/Event.hh"
#include "Rivet/Projections/FastJets.hh"


namespace Rivet {

  class PseudoBoostedTop : public FinalState {
  public:
    // Default constructor
    PseudoBoostedTop (double lepMinPt = 45., double lepMaxEta = 2.1,
                      double jetMinPt = 30., double jetMaxEta = 2.4,
                      double bjetR = 0.5, double tjetR = 0.8) :
      FinalState(-MAXDOUBLE, MAXDOUBLE, 0*GeV),
      _lepMinPt(lepMinPt), _lepMaxEta(lepMaxEta),
      _jetMinPt(jetMinPt), _jetMaxEta(jetMaxEta),
      _bjetR(bjetR), _tjetR(tjetR)
    {
      setName("PseudoBoostedTop");
    }

    enum DecayMode {CH_HADRON, CH_MUON, CH_ELECTRON, CH_TAU};
    enum TTbarMode {CH_DILEP, CH_ALLHADRON, CH_SEMILEP_MU, CH_SEMILEP_EL, CH_SEMILEP_TAU};

    TTbarMode mode() const {
      if ((_topDecay == CH_MUON   && _antitopDecay == CH_HADRON) ||
          (_topDecay == CH_HADRON && _antitopDecay == CH_MUON)) {
        return CH_SEMILEP_MU;
      } else if ((_topDecay == CH_ELECTRON && _antitopDecay == CH_HADRON) ||
                 (_topDecay == CH_HADRON   && _antitopDecay == CH_ELECTRON)) {
        return CH_SEMILEP_EL;
      } else if ((_topDecay == CH_TAU    && _antitopDecay == CH_HADRON) ||
                 (_topDecay == CH_HADRON && _antitopDecay == CH_TAU)) {
        return CH_SEMILEP_TAU;
      } else if (_topDecay == CH_HADRON && _antitopDecay == CH_HADRON) {
        return CH_ALLHADRON;
      } else {
        return CH_DILEP;
      }
    }

  public:
    Particle partonTop() const   {
      if (_antitopDecay == CH_HADRON) return _antitop;
      else if (_topDecay == CH_HADRON) return _top;
      else return nullptr;
    }
    Jet particleTop() const {return _particleTop;}

    bool isGenTopJet() const {return _isGenTopJet;}
    bool passParticle() const {return _passParticle;}

  protected:
    bool isFromTop(const GenParticle* p);
    double getDeltaR(const FourMomentum& p1, const FourMomentum& p2);
    bool fromHardMuon(const GenParticle* p);

  private:
    const double _lepMinPt, _lepMaxEta;
    const double _jetMinPt, _jetMaxEta;
    const double _bjetR, _tjetR;

    constexpr static double _halfpi = 3.1415 / 2.;

  private:
    bool _isValid;

    DecayMode _topDecay;
    DecayMode _antitopDecay;

    Particle _top;
    Particle _antitop;

    Jet _particleTop;

    bool _isParticleLep;
    bool _isGenBJet;
    bool _isGenTopJet;
    bool _passParticle;

    float _mtt;
  };

  struct GreaterByPt
  {
    bool operator() (const Particle& a, const Particle& b) {
      return a.pt() > b.pt();
    }
  };

  struct GreaterByPtJet
  {
    bool operator() (const Jet& a, const Jet& b) {
      return a.momentum().pt() > b.momentum().pt();
    }
  };

  // PseudoBoostedTop member function definitions
  double PseudoBoostedTop::getDeltaR(const FourMomentum& p1, const FourMomentum& p2) {
    double deta = std::fabs(p1.eta() - p2.eta());
    double dphi = std::fabs(p1.phi() - p2.phi());
    return std::sqrt(deta * deta + dphi * dphi);
  }

  // Determine if lepton comes from hard top decay
  bool PseudoBoostedTop::isFromTop(const GenParticle* p) {
    bool fromTop = false;
    const int pdgId = p->pdg_id();
    GenVertex* prodVtx = p->production_vertex();
    if (prodVtx != nullptr) {
      foreach (const GenParticle* p1, Rivet::particles(prodVtx, HepMC::parents)) {
        const int pdgIdParent = p1->pdg_id();
        if (pdgIdParent == 24) { // parent is a W+ --> leptonic top
          if (pdgId == -11) _topDecay = CH_ELECTRON;
          if (pdgId == -13) _topDecay = CH_MUON;
          if (pdgId == -15) _topDecay = CH_TAU;
          fromTop = true;
        }
        else if (pdgIdParent == -24) { // parent is a W- --> leptonic antitop
          if (pdgId == 11) _antitopDecay = CH_ELECTRON;
          if (pdgId == 13) _antitopDecay = CH_MUON;
          if (pdgId == 15) _antitopDecay = CH_TAU;
          fromTop = true;
        }
      }
    }
    return fromTop;
  }

  bool PseudoBoostedTop::fromHardMuon(const GenParticle* p) {
    GenVertex* prodVtx = p->production_vertex();
    if (prodVtx == nullptr) return false;
    foreach (const GenParticle* ancestor, Rivet::particles(prodVtx, HepMC::ancestors)) {
      if (std::abs(ancestor->pdg_id()) == 13 && ancestor->status() == 3)
        return true;
    }
    return false;
  }


  // Rivet plugin
  class CMS_2015_I1388555 : public Analysis {
  public:

    // Constructor
    CMS_2015_I1388555() : Analysis("CMS_2015_I1388555") {}

    // Book histograms and initialise projections before the run
    void init() {
      addProjection(PseudoBoostedTop(45., 2.1, 30., 2.4, 0.5, 0.8), "ttbar");

      _hEl_topPt_parton          = bookHisto1D("d01-x01-y01"); // dsigma/dpt(top)
      _hEl_topPt_particle        = bookHisto1D("d02-x01-y01"); // dsigma/dpt(top)
      _hEl_topY_parton           = bookHisto1D("d03-x01-y01"); // dsigma/dy(top)
      _hEl_topY_particle         = bookHisto1D("d04-x01-y01"); // dsigma/dy(top)

      _hMu_topPt_parton          = bookHisto1D("d05-x01-y01"); // dsigma/dpt(top)
      _hMu_topPt_particle        = bookHisto1D("d06-x01-y01"); // dsigma/dpt(top)
      _hMu_topY_parton           = bookHisto1D("d07-x01-y01"); // dsigma/dy(top)
      _hMu_topY_particle         = bookHisto1D("d08-x01-y01"); // dsigma/dy(top)

      _hComb_topPt_parton        = bookHisto1D("d09-x01-y01"); // dsigma/dpt(top)
      _hComb_topPt_particle      = bookHisto1D("d10-x01-y01"); // dsigma/dpt(top)
      _hComb_topY_parton         = bookHisto1D("d11-x01-y01"); // dsigma/dy(top)
      _hComb_topY_particle       = bookHisto1D("d12-x01-y01"); // dsigma/dy(top)

      _hEl_topPt_parton_norm     = bookHisto1D("d13-x01-y01"); // 1/sigma dsigma/dpt(top)
      _hEl_topPt_particle_norm   = bookHisto1D("d14-x01-y01"); // 1/sigma dsigma/dpt(top)
      _hEl_topY_parton_norm      = bookHisto1D("d15-x01-y01"); // 1/sigma dsigma/dy(top)
      _hEl_topY_particle_norm    = bookHisto1D("d16-x01-y01"); // 1/sigma dsigma/dy(top)

      _hMu_topPt_parton_norm     = bookHisto1D("d17-x01-y01"); // 1/sigma dsigma/dpt(top)
      _hMu_topPt_particle_norm   = bookHisto1D("d18-x01-y01"); // 1/sigma dsigma/dpt(top)
      _hMu_topY_parton_norm      = bookHisto1D("d19-x01-y01"); // 1/sigma dsigma/dy(top)
      _hMu_topY_particle_norm    = bookHisto1D("d20-x01-y01"); // 1/sigma dsigma/dy(top)

      _hComb_topPt_parton_norm   = bookHisto1D("d21-x01-y01"); // 1/sigma dsigma/dpt(top)
      _hComb_topPt_particle_norm = bookHisto1D("d22-x01-y01"); // 1/sigma dsigma/dpt(top)
      _hComb_topY_parton_norm    = bookHisto1D("d23-x01-y01"); // 1/sigma dsigma/dy(top)
      _hComb_topY_particle_norm  = bookHisto1D("d24-x01-y01"); // 1/sigma dsigma/dy(top)

      nMu = 0;
      nEl = 0;
      nPassParton_mu = 0;
      nPassParticle_mu = 0;
      nPassParton_el = 0;
      nPassParticle_el = 0;
    }


    /// Perform the per-event analysis
    void analyze(const Event& event) {
      const double weight = event.weight();

      // Get the ttbar candidate
      const PseudoBoostedTop& ttbar = applyProjection<PseudoBoostedTop>(event, "ttbar");

      if (ttbar.mode() == PseudoBoostedTop::CH_DILEP ||
          ttbar.mode() == PseudoBoostedTop::CH_ALLHADRON ||
          ttbar.mode() == PseudoBoostedTop::CH_SEMILEP_TAU) vetoEvent;

      // Do selection
      const FourMomentum& partonTopP4 = ttbar.partonTop().momentum();

      if (ttbar.mode() == PseudoBoostedTop::CH_SEMILEP_MU) {
        // pass loose parton, muon channel
        nMu += 1;
        _hMu_topPt_parton->fill(partonTopP4.pT(), weight);
        _hMu_topPt_parton_norm->fill(partonTopP4.pT(), weight);
        _hComb_topPt_parton->fill(partonTopP4.pT(), weight);
        _hComb_topPt_parton_norm->fill(partonTopP4.pT(), weight);

        if (partonTopP4.pT() >= 400.) {
          nPassParton_mu += 1;
          _hMu_topY_parton->fill(partonTopP4.rapidity(), weight);
          _hMu_topY_parton_norm->fill(partonTopP4.rapidity(), weight);
          _hComb_topY_parton->fill(partonTopP4.rapidity(), weight);
          _hComb_topY_parton_norm->fill(partonTopP4.rapidity(), weight);
        }

        if (ttbar.isGenTopJet()) {
          const FourMomentum& particleTopP4 = ttbar.particleTop().momentum();

          _hMu_topPt_particle->fill(particleTopP4.pT(), weight);
          _hMu_topPt_particle_norm->fill(particleTopP4.pT(), weight);
          _hComb_topPt_particle->fill(particleTopP4.pT(), weight);
          _hComb_topPt_particle_norm->fill(particleTopP4.pT(), weight);

          if (ttbar.passParticle()){
            nPassParticle_mu += 1;
            _hMu_topY_particle->fill(particleTopP4.rapidity(), weight);
            _hMu_topY_particle_norm->fill(particleTopP4.rapidity(), weight);
            _hComb_topY_particle->fill(particleTopP4.rapidity(), weight);
            _hComb_topY_particle_norm->fill(particleTopP4.rapidity(), weight);
          }
        }
      }

      else if (ttbar.mode() == PseudoBoostedTop::CH_SEMILEP_EL) {
        nEl += 1;
        _hEl_topPt_parton->fill(partonTopP4.pT(), weight);
        _hEl_topPt_parton_norm->fill(partonTopP4.pT(), weight);
        _hComb_topPt_parton->fill(partonTopP4.pT(), weight);
        _hComb_topPt_parton_norm->fill(partonTopP4.pT(), weight);

        if (partonTopP4.pT() >= 400.) {
          nPassParton_el += 1;
          _hEl_topY_parton->fill(partonTopP4.rapidity(), weight);
          _hEl_topY_parton_norm->fill(partonTopP4.rapidity(), weight);
          _hComb_topY_parton->fill(partonTopP4.rapidity(), weight);
          _hComb_topY_parton_norm->fill(partonTopP4.rapidity(), weight);
        }

        if (ttbar.isGenTopJet()) {
          const FourMomentum& particleTopP4 = ttbar.particleTop().momentum();

          _hEl_topPt_particle->fill(particleTopP4.pT(), weight);
          _hEl_topPt_particle_norm->fill(particleTopP4.pT(), weight);
          _hComb_topPt_particle->fill(particleTopP4.pT(), weight);
          _hComb_topPt_particle_norm->fill(particleTopP4.pT(), weight);

          if (ttbar.passParticle()) {
            nPassParticle_el += 1;
            _hEl_topY_particle->fill(particleTopP4.rapidity(), weight);
            _hEl_topY_particle_norm->fill(particleTopP4.rapidity(), weight);
            _hComb_topY_particle->fill(particleTopP4.rapidity(), weight);
            _hComb_topY_particle_norm->fill(particleTopP4.rapidity(), weight);
          }
        }
      }

    }


    // Normalise histograms etc., after the run
    void finalize() {
      double xs_mu_parton = 252.89 * 1000. * nPassParton_mu / nMu;
      double xs_mu_particle = 252.89 * 1000. * nPassParticle_mu / nMu;
      double xs_el_parton = 252.89 * 1000. * nPassParton_el / nEl;
      double xs_el_particle = 252.89 * 1000. * nPassParticle_el / nEl;
      double xs_comb_parton = 252.89 * 1000. * (nPassParton_el + nPassParton_mu) / (nEl + nMu);
      double xs_comb_particle = 252.89 * 1000. * (nPassParticle_el + nPassParton_mu) / (nEl + nMu);

      normalize(_hMu_topPt_parton,     xs_mu_parton,     false);
      normalize(_hMu_topPt_particle,   xs_mu_particle,   false);
      normalize(_hMu_topY_parton,      xs_mu_parton,     false);
      normalize(_hMu_topY_particle,    xs_mu_particle,   false);
      normalize(_hEl_topPt_parton,     xs_el_parton,     false);
      normalize(_hEl_topPt_particle,   xs_el_particle,   false);
      normalize(_hEl_topY_parton,      xs_el_parton,     false);
      normalize(_hEl_topY_particle,    xs_el_particle,   false);
      normalize(_hComb_topPt_parton,   xs_comb_parton,   false);
      normalize(_hComb_topPt_particle, xs_comb_particle, false);
      normalize(_hComb_topY_parton,    xs_comb_parton,   false);
      normalize(_hComb_topY_particle,  xs_comb_particle, false);

      normalize(_hMu_topPt_parton_norm,     1.0, false);
      normalize(_hMu_topPt_particle_norm,   1.0, false);
      normalize(_hMu_topY_parton_norm,      1.0, false);
      normalize(_hMu_topY_particle_norm,    1.0, false);
      normalize(_hEl_topPt_parton_norm,     1.0, false);
      normalize(_hEl_topPt_particle_norm,   1.0, false);
      normalize(_hEl_topY_parton_norm,      1.0, false);
      normalize(_hEl_topY_particle_norm,    1.0, false);
      normalize(_hComb_topPt_parton_norm,   1.0, false);
      normalize(_hComb_topPt_particle_norm, 1.0, false);
      normalize(_hComb_topY_parton_norm,    1.0, false);
      normalize(_hComb_topY_particle_norm,  1.0, false);
    }


  private:
    Histo1DPtr _hMu_topPt_parton;
    Histo1DPtr _hMu_topPt_particle;
    Histo1DPtr _hMu_topY_parton;
    Histo1DPtr _hMu_topY_particle;
    Histo1DPtr _hEl_topPt_parton;
    Histo1DPtr _hEl_topPt_particle;
    Histo1DPtr _hEl_topY_parton;
    Histo1DPtr _hEl_topY_particle;
    Histo1DPtr _hComb_topPt_parton;
    Histo1DPtr _hComb_topPt_particle;
    Histo1DPtr _hComb_topY_parton;
    Histo1DPtr _hComb_topY_particle;
    Histo1DPtr _hMu_topPt_parton_norm;
    Histo1DPtr _hMu_topPt_particle_norm;
    Histo1DPtr _hMu_topY_parton_norm;
    Histo1DPtr _hMu_topY_particle_norm;
    Histo1DPtr _hEl_topPt_parton_norm;
    Histo1DPtr _hEl_topPt_particle_norm;
    Histo1DPtr _hEl_topY_parton_norm;
    Histo1DPtr _hEl_topY_particle_norm;
    Histo1DPtr _hComb_topPt_parton_norm;
    Histo1DPtr _hComb_topPt_particle_norm;
    Histo1DPtr _hComb_topY_parton_norm;
    Histo1DPtr _hComb_topY_particle_norm;

    int nMu;
    int nEl;
    int nPassParton_mu;
    int nPassParticle_mu;
    int nPassParton_el;
    int nPassParticle_el;
  };


  // The hook for the plugin system
  DECLARE_RIVET_PLUGIN(CMS_2015_I1388555);
}
